#! /usr/bin/perl -w

# Minimum number of lines of context gdb can produce
$context = 0;

# The location of the debauch library
$debauchlib = "/usr/lib/libdebauch.so";

$program = shift @ARGV;
$args = "";

$debauchoutput = "leaks.out";

system("LD_PRELOAD=$debauchlib $program $args 2> $debauchoutput");

## Save all addresses that come up in the stack backtrace into
## %funcaddress.
open(DEBAUCHOUTPUT, $debauchoutput);
while (<DEBAUCHOUTPUT>) {
  if (/return stack:\s+(.*)/) {
    @addresslist = split(/ /, $1);
    foreach $address (@addresslist) {
      $funcaddress{$address} = $address;
    }
  }
}

open(GDBSCRIPT, ">gdbscript");
print GDBSCRIPT "set env LD_PRELOAD $debauchlib\n";
print GDBSCRIPT "set listsize $context\n";
print GDBSCRIPT "break main\n";
print GDBSCRIPT "run\n";
foreach $address (keys(%funcaddress)) {
  print GDBSCRIPT "list \*$funcaddress{$address}\n";
}
print GDBSCRIPT "quit\n";
close(GDBSCRIPT);

open(GDBOUTPUT, "gdb $program < gdbscript 2>&1 |");
while (<GDBOUTPUT>) {
  if (/(\S+) is in (\S+) \((\S+)\)/) {
    $address = $1;
    $function = $2;
    $filepos = $3;

    $funcaddress{$address} = "$address   $function() at $filepos";
  }
}
close(GDBOUTPUT);


## Go over the output from debauch and substitute address information.
seek DEBAUCHOUTPUT, 0, 0;
while (<DEBAUCHOUTPUT>) {
  if (/(.*)Saved return stack:\s+(.*)/) {
    $info = $1;
    $stacktrace = $2;

    print "$info\n";
    print "return stack:\n";
    @addresslist = split(/ /, $stacktrace);
    $nr = 0;
    foreach $address (@addresslist) {
      print "\#$nr $funcaddress{$address}\n";
      $nr++;
    }
  } else {
    print;
  }
}
